#include "ext.h"
#include "z_dsp.h"
#include <math.h>
#include "fft_mayer.proto.h"

/******************* PITCHTRACK *********************/

#define NAMPCOEF 8
#define MINPOINTS 128
#define MAXPOINTS 512
#define MAXLOG 9
#define FMAXPOINTS 512.

#define PEROCTAVE 24
#define FPEROCTAVE 24.
#define PERHALFTONE 2
#define UCALL_PAGE_SIZE 0
#define MAXVS 256

#define NKERNEL (PEROCTAVE * MAXLOG - 2 * PEROCTAVE)
#define FREQSTEP 1.0293022
#define LOGSTEP .028881096
//#define PI 3.14159265 // defined in msp headers with better precision
//#define TWOPI (2. * PI) // defined in msp headers with better precision
#define CUTOFF .05
#define ONQUALITYTHRESH .5
#define OFFQUALITYTHRESH .25
#define ONPOWTHRESH .01
#define OFFPOWTHRESH .005
#define STICK .3
#define OVERLAP 2.

#define EXTRAKERNEL 73		/* number of extra filter kernels to keep at end */
#define ALIGN 16			/* byte alignment for microcode */
#define LOWINDEX 6
#define DEFREATTACK .5
#define DEFRATIME 200
#define DEFDEBOUNCETIME 150

#define DUMTAB1SIZE 256
#define DUMTAB2SIZE 1024

static float dumbtab1[DUMTAB1SIZE], dumbtab2[DUMTAB2SIZE];

typedef struct kernel
{
	float *k_coef;
	long k_ncoef;
	long k_onset;
	float k_powcorrect;
} t_kernel;

typedef struct sigpt
{
	t_pxobject x_obj;
	void *x_clock;
	void *x_out2;
	float x_freq;
	float *x_buf;
	float *x_ebuf;
	float *x_fillbuf;
	long x_npts;
	long x_nchan;
	float x_lofreq;
	float x_reffreq;		/* frequency corresponding to MIDI 69 */
	float x_stretchcoef;	/* inverse natural log of a MIDI halftone */
	float x_onpowthresh;
	float x_offpowthresh;
	float x_onqualitythresh;
	float x_offqualitythresh;
	float x_stick;
	float x_ampcoef[NAMPCOEF];
	long x_glisstime;
	float x_halftonesperframe;
	float x_srate;
	float x_lastpitch;
	float x_powslope;
	long x_powtime;
	float x_powdecay;
	float x_powslopethresh;
	float x_lastpow;
	long x_lopitch;
	long x_hipitch;
	long x_lopoint;
	long x_hipoint;
	long x_lastindex;
	long x_midipit;
	long x_overlap;
	float *x_freeme;
	long x_printme;
	long x_peaked;
	long x_midilastout;
	long x_miditime;
	long x_debouncetime;
	long x_downsample;
	float x_maxerror;
#ifdef FTS
	FFT_NET *x_net;		// preallocate FFT memory in put routine
#endif
	char x_fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE];	// moved here to
			// try to avoid running out of stack
} t_sigpt;

double ftom(double f);
double mtof(double f);
float pt_atom_getfloat(t_atom *at);
double dumsqrt(double d);
void pt__init(void);
void sigpt__makespect(float *fftbuf, t_kernel *kp, float *spect, long n, long npts);
t_int *sigpt_perform(t_int *w);
void sigpt_doit(t_sigpt *x);
void sigpt_downsample(float *in, float *out, long np, long ds);
void sigpt_dsp(t_sigpt *x, t_signal **sp);
void sigpt_bang(t_sigpt *x);
void sigpt_pt_tune(t_sigpt *x, double f);
void sigpt_stretch(t_sigpt *x, double f);
void sigpt_ampl_threshold(t_sigpt *x, double f);
void sigpt_quality_threshold(t_sigpt *x, double f);
void sigpt_amp_range(t_sigpt *x, double f1, double f2);
void sigpt_quality_range(t_sigpt *x, double f1, double f2);
void sigpt_pitch_range(t_sigpt *x, long n1, long n2);
void sigpt_fixpitch(t_sigpt *x);
void sigpt_stick(t_sigpt *x, double f);
void sigpt_maxerror(t_sigpt *x, double f);
void sigpt_coefs(t_sigpt *x, t_symbol *s, long ac, t_atom *av);
void sigpt_glisstime(t_sigpt *x, long n);
void sigpt_debouncetime(t_sigpt *x, long n);
void sigpt_setdownsample(t_sigpt *x, long n);
void sigpt_reattack(t_sigpt *x, double f, long n);
void sigpt_print(t_sigpt *x, long n);
void sigpt_tick(t_sigpt *x);
void sigpt_clean(t_sigpt *x);
void *sigpt_new(long npoints, long period);
//void cu_mux(float *in, float *out, long step, long n);
long ilog2(long n);

float sigpt_defcoef[] =
{ 1., .9, .8, .7, .7, .6, .6, .6};

t_kernel *pt_kernel = 0;

void *sigpt_class;

int errno;

void main(void)
{
	setup((t_messlist **)&sigpt_class, (method)sigpt_new, (method)sigpt_clean, (short)sizeof(t_sigpt), 0L, 
		A_DEFLONG, A_DEFLONG, 0);
	addmess((method)sigpt_dsp, "dsp" , A_CANT, 0);
	addbang((method)sigpt_bang);
	addmess((method)sigpt_pt_tune, "pt_tune", A_FLOAT, 0);
	addmess((method)sigpt_ampl_threshold, "ampl_threshold", A_FLOAT, 0);
	addmess((method)sigpt_quality_threshold, "quality_threshold",
			A_FLOAT, 0);
	addmess((method)sigpt_pt_tune, "reference-pitch", A_FLOAT, 0);
	addmess((method)sigpt_maxerror, "max-error", A_FLOAT, 0);
	addmess((method)sigpt_stretch, "stretch", A_FLOAT, 0);
	addmess((method)sigpt_amp_range, "amp-range", A_FLOAT, A_FLOAT, 0);
	addmess((method)sigpt_quality_range, "quality-range",
			A_FLOAT, A_FLOAT, 0);
	addmess((method)sigpt_pitch_range, "pitch-range", A_LONG, A_LONG, 0);
	addmess((method)sigpt_stick, "stick", A_FLOAT, 0);
	addmess((method)sigpt_coefs, "coefs", A_GIMME, 0);
	addmess((method)sigpt_glisstime, "gliss-time", A_LONG, 0);
	addmess((method)sigpt_debouncetime, "debounce-time", A_LONG, 0);
	addmess((method)sigpt_reattack, "reattack", A_FLOAT, A_LONG, 0);
	addmess((method)sigpt_print, "print", A_DEFLONG, 0);
	addmess((method)sigpt_setdownsample, "downsample", A_LONG, 0);
	//u_inletmethod(0);
	dsp_initclass();
}


long ilog2(long n)
{
	long ret = -1;
	while (n) {
		n >>= 1;
		ret++;
	}
	return (ret);
}

#ifdef FTS
void cu_mux(float *in, float *out, long step, long n)
{
	while (n--) {
		*out = *in++;
		out += step;
	}
}
#endif

double ftom(double f)
{
	if (f > 0)
		return (69. + (1./.057762265) * log(f/440.));
	else
		return (-1000);
}

double mtof(double f)
{
	return (440. * exp(.057762265 * (f - 69.)));
}

float pt_atom_getfloat(t_atom *at)
{
	float f;
	if (at->a_type == A_LONG) 
		f = at->a_w.w_long;
	else if (at->a_type == A_FLOAT)
		f = at->a_w.w_float;
	else if (sscanf(at->a_w.w_sym->s_name, "%f", &f) < 1) {
		error("%s: bad number", at->a_w.w_sym->s_name);
		f = 0;
	}
	return f;
}

double dumsqrt(double d)
{
	float f = d;
	long l = *(long *)(&f);
	
	if (f < 0)
		return (0.);
	else
		return (dumbtab1[(l >> 23) & 0xff] * dumbtab2[(l >> 13) & 0x3ff]);
}

void pt__init(void)
{
	float *specspace, *espec;
	/* estimate total size of kernels: cover the whole spectrum 
		 4 times and overlap 4 points apiece */
	long specsize = 4 * NKERNEL + OVERLAP * MAXPOINTS, filtno;
	long i;
	float *fftbuf, filtfreq, filtomega;
	t_kernel *kp;
	char fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE];
#ifdef FTS
	FFT_NET *net;
#endif
	
	if (pt_kernel)		// already inited
		return;
	if (!(specspace = (float *)sysmem_newptr(specsize * sizeof(*specspace)))) {
		error("pitchtrack~: out of memory");
		return;
	}
	espec = specspace + specsize;
	if (!(pt_kernel = (t_kernel *)sysmem_newptr(NKERNEL * sizeof(*pt_kernel)))) {
		sysmem_freeptr((char *)specspace);
		error("pitchtrack~: out of memory");
		return;
	}
	/* get a page-aligned buffer for doing FFT */
	//fftbuf = (float *)(((long)fftspace) +
	//			 ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
	fftbuf = (float *)fftspace;
#ifdef FTS
	net = fft_setup(MAXPOINTS,RECTANGULAR);
#endif
	for (i = 0, filtfreq = 2. / FMAXPOINTS, kp = pt_kernel; i < NKERNEL; 
			 i++, filtfreq *= FREQSTEP, kp++) {
		float protowind = 1. / (OVERLAP * (FREQSTEP - 1.) * filtfreq),
			window, phase, wphase,
			filtomega = filtfreq * TWOPI, windomega, cutval, norm, midipit;
		float *fp1, *fp2, *fp3, *fp4;
		long nbin;
		long j;
		
		window = (protowind > .5 * FMAXPOINTS ? .5 * FMAXPOINTS : protowind);
		windomega = PI / window;
		set_zero(fftbuf, 2 * MAXPOINTS);
		norm = 1./window;
#ifdef FTS
		for (fp1 = fp2 = fftbuf + MAXPOINTS, j = window, phase = 0., wphase = 0.; j--;
	 		fp1 -= 2, fp2 += 2, phase += filtomega, wphase += windomega) {
			double wind = norm * (.5 + .5 * cos(wphase));
			fp1[0] = fp2[0] =	cos(phase) * wind;
			fp1[1] = -(fp2[1] =	sin(phase) * wind);
#else
		for (fp1 = fp2 = fftbuf + (MAXPOINTS/2), 
			fp3 = fp4 = fftbuf + MAXPOINTS + (MAXPOINTS/2), j = window, phase = 0., wphase = 0.; j--;
	 		fp1--, fp3--, fp2++, fp4++, phase += filtomega, wphase += windomega) {
			double wind = norm * (.5 + .5 * cos(wphase));
			fp1[0] = fp2[0] =	cos(phase) * wind;
			fp3[0] = -(fp4[0] =	sin(phase) * wind);
#endif



			/* if (!i && j > window-3)
				 {
				 post("1 = %f", (double)1.);
				 post("window %f", window);
				 post("1./window %f", 1./window);
				 post("norm	%f", norm);
				 post("norm > .0001 %d", norm > .0001);
				 post("windomega %f", windomega);
				 } */
		}
					/* if (i == 0)
		{
		float *fp3;
		long z = 10;
		for (z = 10, fp3 = fftbuf + MAXPOINTS + 2; z -- > 0; fp3 += 2)
		post("%f %f", fp3[0], fp3[1]);
		post("at end:");
		for (z = 10, fp3 = fftbuf + (2 * MAXPOINTS - 20);
		z -- > 0; fp3 += 2)
		post("%f %f", fp3[0], fp3[1]);
		} */
							
#ifdef FTS
		fft(net,fftbuf, MAXPOINTS, 0);
#else
		fft(MAXPOINTS,fftbuf,fftbuf + MAXPOINTS);
#endif		
		/* if (i == 0)
		{
		float *fp3;
		long z;
		post("---------");
		for (z = 10, fp3 = fftbuf; z -- > 0; fp3 += 2)
		post("%f %f", fp3[0], fp3[1]);
		} */
		
		cutval = CUTOFF;
		for (j = 0, fp2 = fftbuf + (MAXPOINTS/2); j < (MAXPOINTS/2); j++, fp2--)
			if (*fp2 >= cutval || *fp2 < (-cutval))
				break;
		for (j = 0, fp1 = fftbuf; j < (MAXPOINTS/2); j++, fp1++)
			if (*fp1 >= cutval || *fp1 < (-cutval))
				break;
		if (fp1 >= fp2) {
			static long bugged = 0;
			if (!bugged) {
				error("pt~ %d: zero-length kernel", i);
				bugged = 1;
			}
			fp1 = fp2;
		}
		nbin = 1 + ((fp2 - fp1));
		
		/* if (i == 100 || i == 101)
		{
		float *fp3;
		long z = 10;
		post("n = %d; start at %d", nbin, (fp1 - fftbuf) >> 1);
		for (z = 10, fp3 = fp1; fp3 <= fp2 && (z -- > 0); fp3 += 2)
		post("%f %f", fp3[0], fp3[1]);
		post("(from zero:");
		for (z = 10, fp3 = fftbuf; z -- > 0; fp3 += 2)
		post("%f %f", fp3[0], fp3[1]);
		} */
		
		if (specspace + nbin >= espec) {
			error("pt~: misestimated the number of bins");
			return;
		}
		kp->k_onset = fp1 - fftbuf;
		kp->k_coef = specspace;
		kp->k_ncoef = nbin;
		kp->k_powcorrect = window / protowind;
			/* post("%d\n", nbin); */
		//post("kernel %ld onset %ld nbin %ld",i,kp->k_onset,kp->k_ncoef);
		for (; nbin--; fp1++)
			*specspace++ = *fp1;
	}
	
	// initialize square root table
	
	for (i = 0; i < DUMTAB1SIZE; i++) {
		float f;
		long l = (i ? (i == DUMTAB1SIZE-1 ? DUMTAB1SIZE-2 : i) : 1)<< 23;
		*(long *)(&f) = l;
		dumbtab1[i] = sqrt(f);	
	}
	for (i = 0; i < DUMTAB2SIZE; i++) {
		float f = 1 + (1./DUMTAB2SIZE) * i;
		dumbtab2[i] = sqrt(f);	
	}
}

void sigpt__makespect(float *fftbuf, t_kernel *kp, float *spect, long n, long npts)
{
	while (n--) {
		float rsum = 0, isum = 0;
		float *rfp = fftbuf + kp->k_onset;
		float *ifp = fftbuf + npts + kp->k_onset;
		float *cp = kp->k_coef;
		long k = kp->k_ncoef;
		while (k--) {
			float coef = *cp++;
			rsum += *rfp++ * coef;
			isum += *ifp++ * coef;
		}
		*spect++ = rsum * rsum + isum * isum;
		kp++;
	}
}

t_int *sigpt_perform(t_int *w)
{
	t_sigpt *x = (t_sigpt *)(w[1]);
	float *in = (float *)(w[2]);
	int n = w[3];
	

	float *fp = x->x_fillbuf;
	t_sigpt *debug = x;
	
	if (x->x_obj.z_disabled)
		goto out;

	sigpt_downsample(in, fp, n, x->x_downsample);
	fp += (n / x->x_downsample);
	if (fp != x->x_ebuf)
		x->x_fillbuf = fp;
	else
		sigpt_doit(x);
out:
	return w + 4;
}

void sigpt_doit(t_sigpt *x)
{
	long npts = x->x_npts, boundary;
	long nchan = x->x_nchan, index = -1, i,
	lo = x->x_lopoint-2, hi = x->x_hipoint;
	float bestamp, pitchpow, totpow, newpitch, *fp, *fp2, *fftbuf;
	t_kernel *kp = pt_kernel;
	
	//char fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE];
	float para[3], correct, spectrum[NKERNEL + EXTRAKERNEL],
		pitches[NKERNEL], ampspect[NKERNEL + EXTRAKERNEL];
		
	/* get a page-aligned buffer for doing FFT */
	//fftbuf = (float *)(((long)fftspace) +
	//			 ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
	boundary = npts;
	fftbuf = (float *)x->x_fftspace;
	set_zero(fftbuf, 2 * npts);		/* zero it out */
#ifdef FTS
	cu_mux(x->x_buf, fftbuf, 2, npts);	/* copy input into real part */
#else
	sysmem_copyptr(x->x_buf,fftbuf, npts * sizeof(float));
#endif
	x->x_fillbuf = x->x_buf + x->x_overlap;
	if (x->x_overlap)
		sysmem_copyptr((char *)(x->x_buf + (npts - x->x_overlap)), (char *)x->x_buf,
			x->x_overlap * sizeof(float));
	
#ifdef FTS
	fft(x->x_net,fftbuf, npts, 0);
#else
	fft(npts,fftbuf,fftbuf + boundary);
#endif
	/* get a bounded-Q power spectrum, quarter-tone spaced */
	sigpt__makespect(fftbuf, pt_kernel, spectrum, nchan, boundary);
	for (i = EXTRAKERNEL, fp = spectrum + NKERNEL; i--;)
		*fp++ = 0;
	
	if (x->x_printme >= 2) {
		long z;
		float *fp3 = spectrum;
		post("raw filter output:");
		for (z = 0; z < nchan; z += 4, fp3 += 4)
			post("%3.3d %12.1f %12.1f %12.1f %12.1f", z, fp3[0], fp3[1], fp3[2], fp3[3]);
	}
	
	/* make an amplitude spectrum */
	
	for (i = NKERNEL, fp = spectrum, fp2 = ampspect; i--;)
//#ifdef I860
		*fp2++ = dumsqrt(*fp++);
//#else
//		*fp2++ = sqrt(*fp++);
//#endif
	for (i = EXTRAKERNEL; i--;)
		*fp2++ = 0;
	
	
	/* look for bin with max. power in (itself and first few
		 overtones.) */
	for (i = 0, fp = ampspect, fp2 = x->x_ampcoef, bestamp=0; i < nchan; fp++, i++) {
		float mofo = fp[0] * fp2[0] + fp[24] * fp2[1] + fp[38] * fp2[2] +
			fp[48] * fp2[3] + fp[56] * fp2[4] + fp[62] * fp2[5] + fp[67] * fp2[6] +
			fp[72] * fp2[7];
		/*
		sigpt_coef[1] * fp[24]
		+ sigpt_coef[2] * fp[38]
		+ sigpt_coef[3] * fp[48]
		+ sigpt_coef[4] * fp[56]
		+ sigpt_coef[5] * fp[62]
		+ sigpt_coef[6] * fp[67]
		+ sigpt_coef[7] * fp[72]; */
		pitches[i] = mofo;
		if (i >= lo && i < hi && mofo > bestamp)
			bestamp = mofo, index = i;
	}
	if (index <= 0)	 {		/* didn't find anything at all? */
		x->x_freq = newpitch = 0;
		x->x_lastindex = 0;
		goto done;
	}
	
	if (x->x_lastindex) {		/* possibly stick to previous pitch */
		long lastindex = x->x_lastindex;
		float x2;
		fp = pitches + lastindex;
		x2 = *fp;
		if (fp[-1] > x2 && lastindex > lo)
			x2 = *--fp, lastindex--;
		else if (fp[1] > x2 && lastindex < hi - 1)
			x2 = *++fp, lastindex++;
		if (x2 > (1. - x->x_stick) * bestamp)
			index = lastindex;
	}
	
	/* get total power for the good overtones */
	fp = spectrum + index;
	pitchpow = fp[0] + fp[24] + fp[38] + fp[48] + fp[56] +
		fp[62] + fp[67] + fp[72];
	
	/* get the total power from DC up to the 8th harmonic.
		 The k_powcorrect factor is to correct for the fact that the
		 filterbank overlap is greater in the low bands and so
		 merely summing them would weight them too highly. */
	fp += 72;
	if (fp >= spectrum + NKERNEL)
		fp = spectrum + (NKERNEL - 1);
	for (fp2 = spectrum, totpow = 0., kp = pt_kernel; fp2 < fp; fp2++, kp++)
		totpow += *fp2 * kp->k_powcorrect;
	
	if (x->x_printme)
		post("index %d, value %f; overall %f", index, (float)sqrt(pitchpow)/npts, 
			(float)sqrt(totpow)/npts);
	
	/* to report a pitch there must be at least "offqualitythresh"
		 times the total power in the good bins. */
		 
	if (pitchpow < x->x_offpowthresh || pitchpow < x->x_offqualitythresh * totpow) {
		x->x_freq = newpitch = 0;
		x->x_lastindex = 0;
		if (x->x_printme)
			post("no pitch (q or power too low)");
		goto done;
	}
	
	/* remember the index for pitch-stick test next time */
	if (pitchpow > x->x_onpowthresh)
		x->x_lastindex = index;
	else
		x->x_lastindex = 0;
	
	/* do a parabolic correction to increase precision */
	for (i = 0, fp = spectrum + index - 1; i <= 3; i++, fp++)
		para[i] = fp[0] + 2.*fp[24] + 3.*fp[38] + 4.*fp[48] +
			5.*fp[56] + 6.*fp[62] + 7.*fp[67] + 8.*fp[72];
	if (para[0] >= para[1])
		correct = -.5;
	else if (para[2] >= para[1])
		correct = .5;
	else {
		double f1 = 1. - para[0]/para[1], f2 = 1. - para[2]/para[1];
		correct = (f1-f2)/(2.*(f1+f2) - 3.*f1*f2);
	}
	/* correct = .5 * (para[2] - para[0])
		 / (2. * para[1] - para[0] - para[2]); */
	
	x->x_freq = x->x_lofreq * exp(LOGSTEP * (index + correct));
	if (x->x_downsample > 1)
		x->x_freq /= (double)x->x_downsample;
	newpitch = 69. + x->x_stretchcoef * log(x->x_freq/x->x_reffreq);
	if (x->x_printme) {
		post("freq %f; pitch %f", x->x_freq, newpitch);
	}
 done:
	if (x->x_midipit) {
		float error, lastpow = x->x_lastpow;
		/* test for turnoff -- if "newpitch" is more than 3/4
		of a halftone away from the current note, (or from an
		octave up or down from it) the current note has stopped
		playing.	This also takes care of the case where
		newpitch == 0. */
		error = newpitch - x->x_midipit;
		if (error > 6)
			error -= 12;
		else if (error < -6)
			error += 12;
		if (error > .75 || error < -.75)
			x->x_midipit = 0;
		
		/* now test for reattack. */
		if (lastpow > totpow) {
			x->x_peaked = 1;
			x->x_powslope = 0;
		} else {
			float growth = x->x_powslope * x->x_powdecay +
				(lastpow > 0 ? (totpow - lastpow)/lastpow : 0);
			if (growth >= x->x_powslopethresh)
				x->x_midipit = 0;
			x->x_powslope = growth;
		}
	}
	x->x_lastpow = totpow;
	
	/* test for turnon.	We must: not have a current pitch;
		 have a candidate pitch; have at least a certain
		 amount of power and "quality"; and, if there's
		 a "glisstime" specified, the current pitch has to
		 agree roughly with that of the last frame. */
	if (!x->x_midipit && newpitch != 0 &&
			pitchpow > x->x_onpowthresh &&
			pitchpow > x->x_onqualitythresh * totpow &&
			(x->x_glisstime == 0 ||
			 (newpitch >= x->x_lastpitch - x->x_halftonesperframe &&
			newpitch <= x->x_lastpitch + x->x_halftonesperframe))) {
		
		long ipit = newpitch + .5;
		float error = newpitch - ipit;
		if (x->x_printme)
			post("error %f", error);
		if (error < x->x_maxerror && error > -x->x_maxerror) {
			long rats = x->x_midilastout;
			if (gettime() > x->x_miditime || (ipit != rats && ipit != rats + 12 && 
				ipit != rats - 12)) {
				
				x->x_miditime = gettime() + x->x_debouncetime;
				x->x_midilastout = ipit;
				x->x_peaked = 0;
				x->x_powslope = 0;
				clock_delay(x->x_clock, 0L);
			}
			x->x_midipit = ipit;
		} else if (x->x_printme)
			post("out of tune");
	}
	x->x_printme = 0;
	x->x_lastpitch = newpitch;
}

void sigpt_downsample(float *in, float *out, long np, long ds)
{
	long i;
	
	for (i = 0; i < np; i+=ds, in += ds) {
		*out++ = *in;
	}
}
		
#ifdef FTS
void sigpt_put(t_sigpt *x, long whether)
{
	if (whether) {
		u_stdout((t_ugen *)x);
		x->x_srate = x->x_io[0]->s_sr;
		sigpt_fixpitch(x);
		sigpt_glisstime(x, x->x_glisstime);
		sigpt_reattack(x, x->x_powslopethresh, x->x_powtime);
		x->x_net = fft_setup(x->x_npts,RECTANGULAR);
		dspchain_addc((d_method)cu_pt, 3, (long)x, (long)x->x_io[0]->s_shit, x->x_io[0]->s_n,
			0L,0L);
		if (u_debug)
			post("sigpt %lx: (%lx)", x, x->x_io[0]->s_shit);
	}
}
#endif

void sigpt_dsp(t_sigpt *x, t_signal **sp)
{
	if (sp[0]->s_n > MAXVS) {
		error("pt~: signal vector size must be %ld or less",(long)MAXVS);
		return;
	}
	x->x_srate = sp[0]->s_sr;
	sigpt_fixpitch(x);
	sigpt_glisstime(x, x->x_glisstime);
	sigpt_reattack(x, x->x_powslopethresh, x->x_powtime);
	dsp_add(sigpt_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}

void sigpt_bang(t_sigpt *x)
{
	outlet_float(x->x_out2, (double)(x->x_freq));
}

void sigpt_pt_tune(t_sigpt *x, double f)
{
	if (f >= 1.)
		x->x_reffreq = mtof(f);
	else
		x->x_reffreq = 440.;
}

void sigpt_stretch(t_sigpt *x, double f)
{
	if (f < .01)
		f = 1;
	x->x_stretchcoef = 12. / (log(2.) * f);
}

void sigpt_ampl_threshold(t_sigpt *x, double f)
{
	post("warning: ampl_threshold: does nothing; use amp-range");
}

void sigpt_quality_threshold(t_sigpt *x, double f)
{
	post("warning: quality_threshold is obsolete; use quality-range");
	x->x_onqualitythresh = f;
	x->x_offqualitythresh = .5*f;
}

void sigpt_amp_range(t_sigpt *x, double f1, double f2)
{
	double scale = x->x_npts * x->x_npts;
	x->x_offpowthresh = f1 * f1 * scale;
	x->x_onpowthresh = f2 * f2 * scale;
}

void sigpt_quality_range(t_sigpt *x, double f1, double f2)
{
	x->x_offqualitythresh = f1 * f1;
	x->x_onqualitythresh = f2 * f2;
}

void sigpt_pitch_range(t_sigpt *x, long n1, long n2)
{
	x->x_lopitch = n1;
	x->x_hipitch = n2;
	sigpt_fixpitch(x);
}

void sigpt_fixpitch(t_sigpt *x)
{
	double pitchoffirst = ftom(x->x_lofreq = 2. * x->x_srate / x->x_npts);
	long offset1 = .5 + 2 * (x->x_lopitch - pitchoffirst);
	long offset2 = .5 + 2 * (x->x_hipitch - pitchoffirst);
	if (offset1 < LOWINDEX)
		offset1 = LOWINDEX;
	if (offset1 >= x->x_nchan)
		offset1 = x->x_nchan-1;
	if (offset2 < offset1)
		offset2 = offset1;
	if (offset2 >= x->x_nchan)
		offset2 = x->x_nchan-1;
	x->x_lopoint = offset1;
	x->x_hipoint = offset2;
}

void sigpt_stick(t_sigpt *x, double f)
{
	if (f < 0.)
		f = 0.;
	else if (f > 1.)
		f = 1.;
	x->x_stick = f;
}

void sigpt_maxerror(t_sigpt *x, double f)
{
	if (f < 0.)
		f = 0.;
	else if (f > .5)
		f = .5;
	x->x_maxerror = f;
}

void sigpt_coefs(t_sigpt *x, t_symbol *s, long ac, t_atom *av)
{
	if (ac < NAMPCOEF)
		error("sigpt_coefs; need %d values", NAMPCOEF);
	else {
		long i;
		for (i = 0; i < NAMPCOEF; i++) {
			float f = pt_atom_getfloat(av+i);
			if (f < 0)
				f = 0;
			else if (f > 1)
				f = 1;
			x->x_ampcoef[i] = f;
		}
	}
}

void sigpt_glisstime(t_sigpt *x, long n)
{
	if (n > 0) {
		x->x_glisstime = n;
		x->x_halftonesperframe = (x->x_npts - x->x_overlap)/(.001 * n * x->x_srate);
	}
	else
		x->x_glisstime = 0;
}

void sigpt_debouncetime(t_sigpt *x, long n)
{
	x->x_debouncetime = (n > 0 ? n : 0);
}

void sigpt_setdownsample(t_sigpt *x, long n)
{
	if (n == 1 || n == 2 || n == 4)
		x->x_downsample = n;
	//post("downsample is now %ld",x->x_downsample);
}

void sigpt_reattack(t_sigpt *x, double f, long n)
{
	if (f > 0 && n > 0) {
		x->x_powslopethresh = f;
		x->x_powtime = (n < 1 ? 1 : n);
		x->x_powdecay = exp(-(x->x_npts - x->x_overlap)/
				(.001 * x->x_powtime * x->x_srate));
			/* post("powdecay = %f", x->x_powdecay); */
	} else {
		x->x_powslopethresh = 0;
		x->x_powtime = 0;
		x->x_powdecay = 0;
	}
}

void sigpt_print(t_sigpt *x, long n)
{
	if (n < 1) n = 1;
	x->x_printme = n;
	post("pitch tracker:");
	post("reference-pitch %f", ftom(x->x_reffreq));
	post("stretch %f", 12. / (log(2.) * x->x_stretchcoef));
	post("stick %f", x->x_stick);
	post("gliss-time %d", x->x_glisstime);
	post("debounce-time %d", x->x_debouncetime);
	post("max-error %f", x->x_maxerror);
	post("reattack %f %d", x->x_powslopethresh, x->x_powtime);
	post("coefs %f %f %f %f",
			 x->x_ampcoef[0],x->x_ampcoef[1],x->x_ampcoef[2],x->x_ampcoef[3]); 
	post("...	 %f %f %f %f",
			 x->x_ampcoef[4],x->x_ampcoef[5],x->x_ampcoef[6],x->x_ampcoef[7]); 
	post("amp-range %f %f",
			 sqrt(x->x_offpowthresh)/x->x_npts,
			 sqrt(x->x_onpowthresh)/x->x_npts);
	post("quality-range %f %f", sqrt(x->x_offqualitythresh),
			 sqrt(x->x_onqualitythresh));
	post("pitch-range %d %d", x->x_lopitch, x->x_hipitch);
	if (x->x_lofreq > 1.)
		{
			post("(possible pitch range: %d - %d)",
		 (long)(ftom(x->x_lofreq) + (.5 + LOWINDEX/2)),
		 (long)(ftom(x->x_lofreq) + (.5 + x->x_nchan/2)));
		}
}

void sigpt_tick(t_sigpt *x)
{
	outlet_int(x->x_obj.z_ob.o_outlet, (long)(x->x_midipit));
}

void sigpt_clean(t_sigpt *x)
{
	freeobject(x->x_clock);
	t_freebytes(x->x_freeme,x->x_npts * sizeof(float) + ALIGN);
	dsp_free((t_pxobject *)x);
}

void *sigpt_new(long npoints, long period)
{
	t_sigpt *x;
	float *fp;
	long logn, i;
	
	if (!npoints)
		npoints = 512;
	logn = ilog2(npoints);
	if (npoints != (1 << logn)) {
		error("pt~: power of two required as argument");
		return (0);
	}
	if (npoints > MAXPOINTS || npoints < MINPOINTS) {
		error("pt~: argument out of range");
		return (0);
	}
	if (period & (MAXVS-1)) {
		error("pt~: period must be a multiple of %d", MAXVS);
		return (0);
	}
	if (period > npoints) {
		error("pt~: period can't exceed window size");
		return (0);
	}
	if (!period)
		period = npoints;
	if (!(fp = (float *)t_getbytes(npoints * sizeof(float) + ALIGN))) {
		error("pt~: out of memory");
		return (0);
	}
	x = (t_sigpt *)newobject(sigpt_class);
	
	x->x_freeme = fp;
	//fp = (float *)(((-(long)fp) & (ALIGN - 1)) + (long)fp);
	set_zero(fp, npoints);
	dsp_setup((t_pxobject *)x,1);
	x->x_out2 = floatout((t_object *)x);
	intout((t_object *)x);
	x->x_clock = clock_new(x, (method)sigpt_tick);
	//fft_init();
	pt__init();
	
	x->x_srate = 44100.;
	x->x_overlap = npoints-period;
	x->x_buf = x->x_fillbuf = fp;
	x->x_ebuf = x->x_buf + npoints;
	x->x_npts = npoints;
	x->x_nchan = PEROCTAVE * logn - 2 * PEROCTAVE;
	x->x_lofreq = 1.;
	x->x_downsample = 1;
	x->x_printme = 0;
	x->x_reffreq = 440.;
	x->x_maxerror = .4;
	sigpt_stretch(x, 1.);
	sigpt_amp_range(x, OFFPOWTHRESH, ONPOWTHRESH);
	sigpt_quality_range(x, OFFQUALITYTHRESH, ONQUALITYTHRESH);
	sigpt_pitch_range(x, 0, 1000);
	sigpt_reattack(x, DEFREATTACK, DEFRATIME);
	sigpt_debouncetime(x, DEFDEBOUNCETIME);
	x->x_midipit = 0;
	x->x_midilastout = x->x_miditime = 0;
	x->x_lastindex = 0;
	x->x_stick = STICK;
	x->x_peaked = 0;
	for (i = 0; i < NAMPCOEF; i++)
		x->x_ampcoef[i] = sigpt_defcoef[i];
	sigpt_glisstime(x, 0);
	return (x);
}

